#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <signal.h>
#include <stdio.h>

#include <libintl.h>
#include <locale.h>
#define PACKAGE "pam_usbguard_astra"
#define LOCALEDIR "/usr/share/locale"
#define _(str) dgettext(PACKAGE, str)
__attribute__((constructor)) void preload(void)
{
       bindtextdomain (PACKAGE, LOCALEDIR);
}

#define PAM_SM_SESSION
#define UNUSED __attribute__ ((unused))

#include <security/pam_modules.h>
#include <security/_pam_macros.h>
#include <security/pam_modutil.h>
#include <security/pam_ext.h>

#include <utmpx.h>
#include <paths.h>

#define CONFIG_USERS "/etc/usbguard-astra/users"
#define CONFIG_COMMON "/etc/usbguard-astra/common"
#define CONFIG_USBGUARD "/etc/usbguard/rules.conf"

#define USBGUARD_LIST "/usr/bin/usbguard generate-policy -P"
#define USBGUARD_RESTART "/bin/systemctl restart usbguard"

#define ALLOW "allow"
// deny and reject rules are ignored yet
#define DENY "block"
#define REJECT "reject"

#define MAX_DEV_STRING 4096

// simple one-way linked-list automation
#define LL_DECLARE(type, name) type *name=NULL, *name##_ptr=NULL, *name##_new;

#define LL_NEW(type, name) { name##_new = (type*)malloc(sizeof(type)); } \

#define LL_ADD(name) { \
if(name##_new) { \
name##_new->next=NULL; \
if(!name) name=name##_new; \
if (name##_ptr) name##_ptr->next=name##_new; \
name##_ptr=name##_new; \
} }

struct devparm
{
	char *name;
	char *value;
	struct devparm *next;
};

struct device
{
	char deny;		// 0 = allow, 1 = deny, 2 = reject
	struct devparm *params;
	struct device *next;
};

struct user
{
	char *name;
	struct device *devices;
	struct user *next;
};

void devparm_free(struct devparm *devparm)
{
	struct devparm *p, *prev;

	for(p = devparm; p;)
	{
		if(p->name)
			free(p->name);

		if(p->value)
			free(p->value);

		prev = p;
		p = p->next;
		free(prev);
	}
}

void device_free(struct device *dev)
{
	struct device *p, *prev;

	for(p = dev; p;)
	{
		devparm_free(p->params);
		prev = p;
		p = p->next;
		free(prev);
	}
}

void user_free(struct user *users)
{
	struct user *u, *prev;

	for(u = users; u; )
	{
		device_free(u->devices);
		prev = u;
		u = u->next;
		free(prev);
	}
}

// returns 0 if devices mismatch by any parameter, 1 otherwise
int dev_match_parm(struct device *d1, struct device *d2)
{
	struct devparm *p1, *p2;

	if(!d1 || !d2)
		return -1;

	for(p1 = d1->params; p1; p1 = p1->next)
	{
		for(p2 = d2->params; p2; p2 = p2->next)
		{
			if(strcmp(p1->name, p2->name))
				continue;
			if(strcmp(p1->value, p2->value))
				return 0;
		}
	}
	return 1;
}

// returns 1 if s is usbguard keyword
int is_usbguard_keyword(char *s)
{
	if (!s) return 0;
	if (!strncmp(s, "all-of", 6)) return 1;
	if (!strncmp(s, "one-of", 6)) return 1;
	if (!strncmp(s, "none-of", 7)) return 1;
	if (!strncmp(s, "equals", 6)) return 1;
	if (!strncmp(s, "equals-ordered", 14)) return 1;
	return 0;
}

int str2dev(char *s, struct device *dev)
{
	int r = -1;
	int i, l, start;
	char cnt = 0;
	struct devparm *p = NULL, *prev = NULL;

	if(!s || !dev)
		return -1;

	dev->params = 0;
	l = strlen(s);
	while(l && s[l - 1] == '\n')
		l--;
	for(i = 0; i < l; i++)
	{
		while(s[i] == ' ' && i < l)
			i++;
		start = i;
		if(s[i] == '"')
		{
			start++;
			i++;
			while(s[i] != '"' && i < l)
				i++;
		}
		else
		{
			while(s[i] != ' ' && i < l)
				i++;
		}

		if(cnt == 0)	// first word - access control
		{
			cnt = 1;
			if(!strncmp(ALLOW, s+start, sizeof(ALLOW) - 1))
				dev->deny = 0;
			else if(!strncmp(DENY, s+start, sizeof(DENY) - 1))
				dev->deny = 1;
			else if(!strncmp(REJECT, s+start, sizeof(REJECT) - 1))
				dev->deny = 2;
			else
				r = -2;

			// ignore non-allow rules yet
			if (dev->deny > 0) r = -2;

			if(r == -2)
				return r;
			dev->params = NULL;
			r = -3;
		}
		else if(cnt == 1)	// name
		{
			p = (struct devparm *) malloc(sizeof(struct devparm));
			if(!p)
				return -1;
			p->next = 0;
			if(prev)
				prev->next = p;

			prev = p;
			if(!dev->params)
				dev->params = p;

			p->name = strndup(s + start, i - start);
			p->value = NULL;
			cnt = 2;
			r = -4;
		}
		else if(cnt == 2)	//value
		{
			if (is_usbguard_keyword(s+start))
			{
				while(s[i] == ' ' && i < l)
					i++;
			}

			if (s[start] == '{' || s[i] == '{') // roll i to closing bracket
			{
				while(s[i] != '}' && i < l)
					i++;
				if (i >= l) // no closing bracket
				{
					r = -5;
					break;
				}
				i++; // take the closing bracket too
			}

			p->value = strndup(s + start, i - start);
			cnt = 1;
			r = 0;
		}
	}

	if(r != 0)
		devparm_free(dev->params);

	return r;
}

int dev2file(struct device *dev, FILE *f)
{
	struct device *d;
	struct devparm *p;
	char *quotes;

	if (!dev) return 0;
	if (!f) return -1;

	for(d = dev; d; d = d->next)
	{
		if (d->deny == 0) fprintf(f, ALLOW);
		else if (d->deny == 1) fprintf(f, DENY);
		else if (d->deny == 2) fprintf(f, REJECT);

		for(p = d->params; p; p = p->next)
		{
			if (!strcmp(p->name, "id") ||
				!strcmp(p->name, "with-interface"))
				quotes="";
			else
				quotes="\"";

			fprintf(f, " %s %s%s%s", p->name, quotes, p->value, quotes);
		}
		fprintf(f, "\n");
	}
	return 0;
}

struct device * load_device_list(FILE *fp)
{
	LL_DECLARE(struct device, dev);
	char devstr[MAX_DEV_STRING];

	while(fgets(devstr, sizeof(devstr) - 1, fp) != NULL)
	{
		LL_NEW(struct device, dev);
		if(!dev_new)
			break;

		if(str2dev(devstr, dev_new) != 0)
		{
			free(dev_new);
			continue;
		}
		LL_ADD(dev);
	}

	return dev;
}

int getconnected(struct device **list)
{
	FILE *fp;

	if(!list)
		return -1;

	fp = popen(USBGUARD_LIST, "r");
	if(fp == NULL)
		return -1;

	*list = load_device_list(fp);

	pclose(fp);

	return 0;
}

struct device *load_user_config(char *username)
{
	char *fname = NULL;
	FILE *conf;
	struct device *out;

	if(!username)
	{
		fname = strdup(CONFIG_COMMON);
	} else
	{
		if(asprintf(&fname, "%s/%s", CONFIG_USERS, username) < 0) return NULL;
	}
	if(!fname) return NULL;

	conf = fopen(fname, "r");
	free(fname);
	if (!conf)
		return NULL;

	out = load_device_list(conf);

	fclose(conf);

	return out;
}

int getusers(struct user **userlist, char *thisuser, int skip)
{
	LL_DECLARE(struct user, users);
	struct user *s;

	struct utmpx data;
	FILE *log_file;
	int r = 0;

	if(!userlist || !thisuser)
		return -1;
	log_file = fopen(_PATH_UTMP, "r");
	if(!log_file)
		return -1;

	while(fread(&data, sizeof(struct utmpx), 1, log_file) == 1)
	{
		// ignore all but logged in users
		if(data.ut_type != 7)
			continue;

		// if this user is already logged in, retcode 1
		// skip this user if skip flag set
		if (!strcmp(data.ut_user, thisuser))
		{
			r = 1;
			if (skip) continue;
		}

		// search for duplicate users
		for(s = users; s; s = s->next)
			if (!strcmp(s->name, data.ut_user)) break;
		if (s) break;

		LL_NEW(struct user, users);
		if(!users_new)
			break;
		users_new->name = strdup(data.ut_user);
		users_new->devices = NULL;
		LL_ADD(users);
	}
	fclose(log_file);

	*userlist = users;
	return r;
}

struct device* combine_devices(struct user *users)
{
	LL_DECLARE(struct device, comb);
	struct device *d, *d1;
	struct user *u, *u1;
	char match;

	if (!users) return NULL;

	// iterate all users
	for(u = users; u; u = u->next)
	{
		if (!u->next && u != users) break; // skip last user, if not only one
		// iterate all devices in user
		for(d = u->devices; d; d = d->next)
		{
			match = 1;
			// iterate all other users, they must have it also
			for(u1 = u->next; u1; u1 = u1->next)
			{
				// iterate other user devices
				for(d1 = u1->devices; d1; d1 = d1->next)
				{
					if (dev_match_parm(d, d1)) // common-allowed
						break;
				}

				if (!d1) // end-of-list reached, no match
					match = 0;
			}

			if(match) // match, output device
			{
				// check if device already present in output
				for(comb_new = comb; comb_new; comb_new = comb_new->next)
					if (dev_match_parm(comb_new, d))
					{
						match = 0;
						break;
					}

				if (match)
				{
					// put device on comb list
					LL_NEW(struct device, comb);
					if(!comb_new)
					{
						device_free(comb);
						return NULL;
					}

					comb_new->deny = d->deny;
					comb_new->params = d->params;
					//  steal dev parameters to avoid double free
					d->params = NULL;

					LL_ADD(comb);
				}
			}
		}
	}
	return comb;
}

int allow_user(struct device *user, struct device *common, struct device *connected, char **bad)
{
	struct device *u, *c;
	int cmn;

	for(c = connected; c; c = c->next)
	{
		cmn = 0;
		for(u = user;;u = u->next)
		{
			// "concatenate" user + common
			if (!u && !cmn)
			{
				cmn = 1;
				u = common;
			}
			if (!u) break;
			
			if(dev_match_parm(u, c)) break;
		}
		// connected device was not found in user+common
		if (!u)
		{
			if (c->params) *bad = c->params->value;
			return 0;
		}
	}
	return 1;
}

int is_admin_group(pam_handle_t *pamh, char *user, char *adm_group)
{
	// mostly copy-pasted from pam_wheel

	struct group *grp;
	char * const *list;
	struct passwd *tpwd = NULL;

	// empty user or group
	if (!user || !adm_group || !adm_group[0])
		return 0;

	tpwd = pam_modutil_getpwnam (pamh, user);
	if (!tpwd)
		return 0;

	grp = pam_modutil_getgrnam (pamh, adm_group);

	// no members in group
	if (!grp || (!grp->gr_mem && (tpwd->pw_gid != grp->gr_gid)))
	{
		pam_syslog(pamh, LOG_NOTICE, "no members in group '%s'", adm_group);
		return 0;
	}

	// primary group
	if (tpwd->pw_gid == grp->gr_gid)
		return 1;

	list = grp->gr_mem;

	while (list && *list) {
		if (strcmp(*list, user) == 0)
			return 1;
		list++;
	}

	return 0;
}

int session_openclose(int open, pam_handle_t * pamh, int flags, int argc, const char **argv)
{
	int ret = PAM_SYSTEM_ERR;
	const void *user = NULL;
	const void *service = NULL;
	char *adm_group = NULL;

	int r;
	struct device *connected = NULL, *common = NULL, *comb = NULL, *newuser = NULL;
	struct user *users = NULL, *u = NULL;
	FILE *out;
	char *baddev = NULL;

	// skip sudo and su services
	r = pam_get_item(pamh, PAM_SERVICE, &service);
	if(r != PAM_SUCCESS || !service)
	{
		pam_syslog(pamh, LOG_NOTICE, "Cannot obtain service name.");
	} else
	{
		if (!strcmp((char*)service, "sudo") ||!strcmp((char*)service, "su"))
		{
			pam_syslog(pamh, LOG_NOTICE, "Skipping checks for sudo/su.");
			return PAM_SUCCESS;
		}
	}

	/* Determine the user name */
	r = pam_get_item(pamh, PAM_USER, &user);
	if(r != PAM_SUCCESS || user == NULL || *(const char *) user == '\0')
	{
		pam_syslog(pamh, LOG_NOTICE, "Cannot obtain the user name.");
		return PAM_USER_UNKNOWN;
	}

	// get admgroup=<group> parameter from argv
	for (; argc-- > 0; ++argv) {
		if (!strncmp(*argv,"admgroup=", 9))
		{
			adm_group = strdup(*argv + 9);
			if (!adm_group) return PAM_SYSTEM_ERR;
		}
	}

	// skip any checks for admin group, if specified
	r = is_admin_group(pamh, (char*)user, adm_group);
	free(adm_group);
	if (r)
	{
		pam_syslog(pamh, LOG_NOTICE, "User in admgroup, no action required.");
		return PAM_SUCCESS;
	}

	r = getusers(&users, (char*)user, !open); // skip current user if session is closing
	if (r == 1) // user found in logged-in userlist
	{
		if (open)
		{
			user_free(users);
			pam_syslog(pamh, LOG_NOTICE, "User already logged in, no action required");
			return PAM_SUCCESS;
		}
	} else if (r != 0)
	{
		pam_syslog(pamh, LOG_WARNING, "Failed to get logged-in users list");
		return PAM_SYSTEM_ERR;
	}

	if (getconnected(&connected) != 0)
	{
		pam_syslog(pamh, LOG_NOTICE, "Can't list connected devices");
		ret = PAM_IGNORE;
		goto end;
	}

	common  = load_user_config(NULL);
	if (open) newuser = load_user_config((char*)user);

	if (!common)
		pam_syslog(pamh, LOG_NOTICE, "Common devices list is empty");

	if (open)
	{
		if (!newuser)
			pam_syslog(pamh, LOG_NOTICE, "User devices list is empty");

		if (!allow_user(newuser, common, connected, &baddev))
		{
			pam_syslog(pamh, LOG_NOTICE, "Devices not allowed for user '%s' are connected (%s)",
				(char*)user, baddev ? baddev : "");

			pam_prompt(pamh, PAM_ERROR_MSG, NULL, _("Devices not allowed for user '%s' are connected (%s)"),
				(char*)user, baddev ? baddev : "");

			ret = PAM_PERM_DENIED;
			goto end;
		}
	}

	if(geteuid() != 0)
	{
		pam_syslog(pamh, LOG_NOTICE, "Skipping modification of usbguard config for non-root");
		ret = PAM_SUCCESS;
		goto end;
	}

	// create ruleset containing intersection of all users' rules
	out = fopen(CONFIG_USBGUARD ".new", "w");
	if (out)
	{
		for(u = users; u; u = u->next)
			u->devices = load_user_config(u->name);

		if (open)
		{
			// attach new user to the userlist
			u = (struct user*)malloc(sizeof(struct user));
			u->name = (char*)user;
			u->devices = newuser;
			u->next = users;
			// avoid double free
			newuser = NULL;
			users = NULL;
		} else
		{
			u = users;
			users = NULL;
		}

		comb = combine_devices(u);

		dev2file(common, out);
		dev2file(comb, out);
		fclose(out);
		chmod(CONFIG_USBGUARD ".new", 0600);
		if(rename(CONFIG_USBGUARD ".new", CONFIG_USBGUARD) == 0)
			system(USBGUARD_RESTART);
		pam_syslog(pamh, LOG_NOTICE, "Usbguard config updated");

	} else pam_syslog(pamh, LOG_WARNING, "Can't write new usbguard config %s", CONFIG_USBGUARD);

	ret = PAM_SUCCESS;
end:
	user_free(u);
	user_free(users);
	device_free(newuser);
	device_free(connected);
	device_free(common);
	device_free(comb);

	return ret;
}


PAM_EXTERN int
pam_sm_open_session(pam_handle_t * pamh, int flags, int argc, const char **argv)
{
	return session_openclose(1, pamh, flags, argc, argv); // 1 = open
}

PAM_EXTERN
	int pam_sm_close_session(pam_handle_t * pamh UNUSED, int flags UNUSED,
				 int argc UNUSED, const char **argv UNUSED)
{
	return session_openclose(0, pamh, flags, argc, argv); // 0 = close
}

#ifdef PAM_STATIC

/* static module data */
struct pam_module _pam_parsec_tmpdir_modstruct = {
	"pam_usbguard_astra",
	NULL,
	NULL,
	NULL,
	pam_sm_open_session,
	pam_sm_close_session,
	NULL,
};

#endif
